@@ -44,6 +44,7 @@ struct _GtkHTML {
 
 	GtkBindingSet        *editor_bindings;
 	GtkWidget            *iframe_parent;
+	HTMLObject           *frame;
 	GtkHTMLEditorAPI     *editor_api;
 	gpointer              editor_data;
 	HTMLEngine           *engine;
@@ -72,9 +73,6 @@ struct _GtkHTML {
 	GtkHTMLPrivate *priv;
 };
 
-/* must be forward referenced *sigh* */
-struct _HTMLEmbedded;
-
 struct _GtkHTMLClass {
 	GtkLayoutClass parent_class;
 	
@@ -100,9 +98,10 @@ struct _GtkHTMLClass {
 	void (* iframe_created)     (GtkHTML *html, GtkHTML *iframe);
 
 	/* keybindings signals */
-	void (* scroll)               (GtkHTML *html, GtkOrientation orientation, GtkScrollType scroll_type, gfloat position);
-	void (* cursor_move)          (GtkHTML *html, GtkDirectionType dir_type, GtkHTMLCursorSkipType skip);
-	void (* command)              (GtkHTML *html, GtkHTMLCommandType com_type);
+	void     (* scroll)               (GtkHTML *html, GtkOrientation orientation, GtkScrollType scroll_type,
+					   gfloat position);
+	void     (* cursor_move)          (GtkHTML *html, GtkDirectionType dir_type, GtkHTMLCursorSkipType skip);
+	gboolean (* command)              (GtkHTML *html, GtkHTMLCommandType com_type);
 
 	/* properties */
 	GtkHTMLClassProperties *properties;
@@ -139,20 +138,24 @@ void       gtk_html_set_editor_api  (GtkHTML          *html,
 
 /* parent iframe setting */
 void       gtk_html_set_iframe_parent       (GtkHTML *html,
-					     GtkWidget *parent);
+					     GtkWidget *parent,
+					     HTMLObject *frame);
 
 /* Debugging.  */
 void  gtk_html_enable_debug  (GtkHTML  *html,
 			      gboolean  debug);
 
 /* Behavior.  */
-void  gtk_html_allow_selection  (GtkHTML  *html,
-				 gboolean  allow);
-void  gtk_html_select_word      (GtkHTML  *html);
-void  gtk_html_select_line      (GtkHTML  *html);
-int   gtk_html_request_paste    (GtkHTML  *html,
-				 gint      type,
-				 gint32    time);
+void  gtk_html_allow_selection            (GtkHTML  *html,
+					   gboolean  allow);
+void  gtk_html_select_word                (GtkHTML  *html);
+void  gtk_html_select_line                (GtkHTML  *html);
+void  gtk_html_select_paragraph           (GtkHTML  *html);
+void  gtk_html_select_paragraph_extended  (GtkHTML  *html);
+void  gtk_html_select_all                 (GtkHTML  *html);
+int   gtk_html_request_paste              (GtkHTML  *html,
+					   gint      type,
+					   gint32    time);
 /* Loading.  */
 GtkHTMLStream *gtk_html_begin             (GtkHTML             *html);
 GtkHTMLStream *gtk_html_begin_content     (GtkHTML             *html,
@@ -232,6 +235,8 @@ void  gtk_html_paste  (GtkHTML *html);
 void  gtk_html_undo  (GtkHTML *html);
 void  gtk_html_redo  (GtkHTML *html);
 
+void  gtk_html_insert_html (GtkHTML *html, const gchar *html_src);
+
 /* misc utils */
 
 void      gtk_html_set_default_background_color  (GtkHTML     *html,
@@ -240,9 +245,14 @@ void      gtk_html_set_default_content_type      (GtkHTML     *html,
 						  gchar       *content_type);
 gpointer  gtk_html_get_object_by_id              (GtkHTML     *html,
 						  const gchar *id);
-void      gtk_html_editor_command                (GtkHTML     *html,
+gboolean  gtk_html_command                       (GtkHTML     *html,
 						  const gchar *command_name);
 gboolean  gtk_html_edit_make_cursor_visible      (GtkHTML     *html);
 
 gboolean  gtk_html_build_with_gconf              (void);
+void      gtk_html_set_magnification             (GtkHTML *html,
+						  gdouble magnification);
+void      gtk_html_zoom_in                       (GtkHTML *html);
+void      gtk_html_zoom_out                      (GtkHTML *html);
+void      gtk_html_zoom_reset                    (GtkHTML *html);
 #endif /* _GTKHTML_H_ */
