@@ -29,6 +29,8 @@
 #include <ctype.h>
 #include <string.h>
 
+#include <gal/widgets/e-unicode.h>
+
 #include "gtkhtml-properties.h"
 
 #include "htmlcolor.h"
@@ -43,6 +45,7 @@
 #include "htmlpainter.h"
 #include "htmlplainpainter.h"
 #include "htmltable.h"
+#include "htmltablecell.h"
 #include "htmltext.h"
 #include "htmltextslave.h"	/* FIXME */
 #include "htmlsearch.h"
@@ -55,6 +58,8 @@ static HTMLClueClass *parent_class = NULL;
 #define HCF_CLASS(x) HTML_CLUEFLOW_CLASS (HTML_OBJECT (x)->klass)
 #define HTML_IS_PLAIN_PAINTER(obj)              (GTK_CHECK_TYPE ((obj), HTML_TYPE_PLAIN_PAINTER))
 
+inline HTMLHAlignType html_clueflow_get_halignment (HTMLClueFlow *flow);
+
 
 static void
 copy (HTMLObject *self,
@@ -73,21 +78,32 @@ get_recursive_length (HTMLObject *self)
 }
 
 static HTMLObject *
-op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
+op_helper (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GList *right, guint *len, gboolean cut)
 {
+	/* if (!from && to && HTML_IS_TABLE (to->data) && to->next && GPOINTER_TO_INT (to->next->data) == 0)
+		return NULL;
+	if (!to && from && HTML_IS_TABLE (from->data) && from->next && GPOINTER_TO_INT (from->next->data) == 1)
+		return NULL; 
+	if (!from && (*len || !(self->prev && HTML_IS_CLUEFLOW (self->prev) && HTML_IS_TABLE (HTML_CLUE (self->prev)->tail))))
+	(*len) ++; */
 	if (!from)
 		(*len) ++;
 	html_clueflow_remove_text_slaves (HTML_CLUEFLOW (self));
-	return (*HTML_OBJECT_CLASS (parent_class)->op_copy) (self, e, from, to, len);
+	return cut
+		? (*HTML_OBJECT_CLASS (parent_class)->op_cut) (self, e, from, to, left, right, len)
+		: (*HTML_OBJECT_CLASS (parent_class)->op_copy) (self, e, from, to, len);
+}
+
+static HTMLObject *
+op_copy (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, guint *len)
+{
+	return op_helper (self, e, from, to, NULL, NULL, len, FALSE);
 }
 
 static HTMLObject *
 op_cut (HTMLObject *self, HTMLEngine *e, GList *from, GList *to, GList *left, GList *right, guint *len)
 {
-	if (!from)
-		(*len) ++;
-	html_clueflow_remove_text_slaves (HTML_CLUEFLOW (self));
-	return (*HTML_OBJECT_CLASS (parent_class)->op_cut) (self, e, from, to, left, right, len);
+	return op_helper (self, e, from, to, left, right, len, TRUE);
 }
 
 static void
@@ -97,6 +113,15 @@ split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, gint lev
 	(*HTML_OBJECT_CLASS (parent_class)->split) (self, e, child, offset, level, left, right);
 }
 
+inline static gboolean
+could_merge (HTMLClueFlow *cf1, HTMLClueFlow *cf2)
+{
+	return ((HTML_IS_TABLE (HTML_CLUE (cf1)->tail) && !html_clueflow_is_empty (cf2))
+		|| (HTML_IS_TABLE (HTML_CLUE (cf2)->head) && !html_clueflow_is_empty (cf1)))
+		? FALSE
+		: TRUE;
+}
+
 static gboolean
 merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList *left, GList *right)
 {
@@ -110,12 +135,13 @@ merge (HTMLObject *self, HTMLObject *with, HTMLEngine *e, GList *left, GList *ri
 
 	printf ("merge flows\n");
 
-	if ((HTML_IS_TABLE (HTML_CLUE (self)->tail) && !html_clueflow_is_empty (cf2))
-	    || (HTML_IS_TABLE (HTML_CLUE (with)->head) && !html_clueflow_is_empty (cf1)))
+	if (!could_merge (cf1, cf2))
 		return FALSE;
 
-	if (html_clueflow_is_empty (cf1))
+	if (html_clueflow_is_empty (cf1)) {
 		cf1->style = cf2->style;
+		html_object_copy_data_from_object (self, with);
+	}
 
 	return (* HTML_OBJECT_CLASS (parent_class)->merge) (self, with, e, left, right);
 }
@@ -326,8 +352,7 @@ calc_min_width (HTMLObject *o,
 	gint w = 0;
 	gboolean add;
 
-	add = (HTML_CLUEFLOW (o)->style == HTML_CLUEFLOW_STYLE_PRE
-	       || HTML_CLUEFLOW (o)->style == HTML_CLUEFLOW_STYLE_NOWRAP);
+	add = HTML_CLUEFLOW (o)->style == HTML_CLUEFLOW_STYLE_PRE;
 
 	cur = HTML_CLUE (o)->head;
 	while (cur) {
@@ -356,20 +381,6 @@ set_line_x (HTMLObject **obj, HTMLObject *run, gint x, gboolean *changed)
 	return x;
 }
 
-static gint
-pref_right_margin (HTMLPainter *p, HTMLClueFlow *clueflow, HTMLObject *o, gint y) 
-{
-	gint fixed_margin = html_object_get_right_margin (o, y);
-
-	/* FIXME: this hack lets us wrap the display at 72 characters when we are using
-	   a plain painter */
-	  
-	if (clueflow->style == HTML_CLUEFLOW_STYLE_PRE || ! HTML_IS_PLAIN_PAINTER(p))
-		return fixed_margin;
-	
-	return MIN (fixed_margin, 72 * html_painter_get_space_width (p, GTK_HTML_FONT_STYLE_SIZE_3, NULL));
-}
-
 /* EP CHECK: should be mostly OK.  */
 /* FIXME: But it's awful.  Too big and ugly.  */
 static gboolean
@@ -378,6 +389,7 @@ calc_size (HTMLObject *o,
 {
 	HTMLVSpace *vspace;
 	HTMLClue *clue;
+	HTMLClueFlow *flow;
 	HTMLObject *obj;
 	HTMLObject *line;
 	HTMLClearType clear;
@@ -401,6 +413,7 @@ calc_size (HTMLObject *o,
 	old_width = o->width;
 
 	clue = HTML_CLUE (o);
+	flow = HTML_CLUEFLOW (o);
 
 	obj = clue->head;
 	line = clue->head;
@@ -414,10 +427,10 @@ calc_size (HTMLObject *o,
 	padding = calc_padding (painter);
 	add_pre_padding (HTML_CLUEFLOW (o), padding);
 
-	lmargin = html_object_get_left_margin (o->parent, o->y);
+	lmargin = html_object_get_left_margin (o->parent, painter, o->y);
 	if (indent > lmargin)
 		lmargin = indent;
-	rmargin = pref_right_margin (painter, HTML_CLUEFLOW (o), o->parent, o->y);
+	rmargin = html_object_get_right_margin (o->parent, painter, o->y);
 
 	w = lmargin;
 	a = 0;
@@ -465,10 +478,9 @@ calc_size (HTMLObject *o,
 						obj->y = o->ascent + obj->ascent + a + d;
 						changed = TRUE;
 					}
-					html_clue_append_left_aligned (HTML_CLUE (o->parent),
-								       HTML_CLUE (c));
+					html_clue_append_left_aligned (HTML_CLUE (o->parent), HTML_CLUE (c));
 
-					lmargin = html_object_get_left_margin (o->parent, o->y);
+					lmargin = html_object_get_left_margin (o->parent, painter, o->y);
 
 					if (indent > lmargin)
 						lmargin = indent;
@@ -487,10 +499,9 @@ calc_size (HTMLObject *o,
 						changed = TRUE;
 					}
 					
-					html_clue_append_right_aligned (HTML_CLUE (o->parent),
-									HTML_CLUE (c));
+					html_clue_append_right_aligned (HTML_CLUE (o->parent), HTML_CLUE (c));
 
-					rmargin = pref_right_margin (painter, HTML_CLUEFLOW (o), o->parent, o->y);
+					rmargin = html_object_get_right_margin (o->parent, painter, o->y);
 				}
 			}
 
@@ -523,7 +534,7 @@ calc_size (HTMLObject *o,
 									    && HTML_OBJECT (HTML_TEXT_SLAVE (run)->owner)
 									    == line
 									    && !HTML_TEXT_SLAVE (run)->posStart),
-							    width_left);
+							    flow->style == HTML_CLUEFLOW_STYLE_PRE ? G_MAXINT : width_left);
 
 				if (fit == HTML_FIT_NONE) {
 					w = set_line_x (&obj, run, w, &changed);
@@ -538,14 +549,14 @@ calc_size (HTMLObject *o,
 
 				/* Algorithm for dealing vertical alignment.
 				   Elements with `HTML_VALIGN_BOTTOM' and
-				   `HTML_VALIGN_CENTER' can be handled
+				   `HTML_VALIGN_MIDDLE' can be handled
 				   immediately.	 Objects with
 				   `HTML_VALIGN_TOP', instead, need to know the
 				   total height of the line, so need to be
 				   handled last.  */
 
 				switch (valign) {
-				case HTML_VALIGN_CENTER: {
+				case HTML_VALIGN_MIDDLE: {
 					gint height;
 					gint half_height;
 
@@ -589,11 +600,8 @@ calc_size (HTMLObject *o,
 
 				html_clue_find_free_area (HTML_CLUE (o->parent),
 							  o->y,
-							  rmargin - lmargin,
-							  a+d,
-							  indent,
-							  &new_y, &new_lmargin,
-							  &new_rmargin);
+							  rmargin - lmargin, a+d, indent, &new_y,
+							  &new_lmargin, &new_rmargin);
 				
 				if (new_y != o->y
 				    || new_lmargin > lmargin
@@ -627,31 +635,33 @@ calc_size (HTMLObject *o,
 					/* we've used up this line so insert a newline */
 					newLine = TRUE;
 				}
-				lmargin = html_object_get_left_margin (o->parent, o->y);
+				lmargin = html_object_get_left_margin (o->parent, painter, o->y);
 				
 				if (indent > lmargin)
 					lmargin = indent;
 
-				rmargin = pref_right_margin (painter, HTML_CLUEFLOW (o), o->parent, o->y);
+				rmargin = html_object_get_right_margin (o->parent, painter, o->y);
 			}
 		}
 		
 		/* if we need a new line, or all objects have been processed
 		   and need to be aligned. */
 		if ( newLine || !obj) {
+			HTMLHAlignType halign;
 			int extra;
 
 			extra = 0;
+			halign = html_clueflow_get_halignment (flow);
 
 			if (w > o->width)
 				o->width = w;
 
-			if (clue->halign == HTML_HALIGN_CENTER) {
+			if (halign == HTML_HALIGN_CENTER) {
 				extra = (rmargin - w) / 2;
 				if (extra < 0)
 					extra = 0;
 			}
-			else if (clue->halign == HTML_HALIGN_RIGHT) {
+			else if (halign == HTML_HALIGN_RIGHT) {
 				extra = rmargin - w;
 				if (extra < 0)
 					extra = 0;
@@ -696,7 +706,7 @@ calc_size (HTMLObject *o,
 					html_object_set_max_ascent (line, painter, a);
 					html_object_set_max_descent (line, painter, d);
 					break;
-				case HTML_VALIGN_CENTER:
+				case HTML_VALIGN_MIDDLE:
 					line->y = o->ascent - a - d + line->ascent;
 					break;
 				case HTML_VALIGN_TOP:
@@ -706,8 +716,8 @@ calc_size (HTMLObject *o,
 					g_assert_not_reached ();
 				}
 
-				if (clue->halign == HTML_HALIGN_CENTER
-				    || clue->halign == HTML_HALIGN_RIGHT)
+				if (halign == HTML_HALIGN_CENTER
+				    || halign == HTML_HALIGN_RIGHT)
 					line->x += extra;
 			}
 
@@ -731,10 +741,10 @@ calc_size (HTMLObject *o,
 
 			o->ascent += o->y - oldy;
 
-			lmargin = html_object_get_left_margin (o->parent, o->y);
+			lmargin = html_object_get_left_margin (o->parent, painter, o->y);
 			if (indent > lmargin)
 				lmargin = indent;
-			rmargin = pref_right_margin (painter, HTML_CLUEFLOW (o), o->parent, o->y);
+			rmargin = html_object_get_right_margin (o->parent, painter, o->y);
 
 			w = lmargin;
 			d = 0;
@@ -990,7 +1000,6 @@ get_tag (HTMLClueFlow *flow)
 	case HTML_CLUEFLOW_STYLE_H6:
 	case HTML_CLUEFLOW_STYLE_ADDRESS:
 	case HTML_CLUEFLOW_STYLE_PRE:
-	case HTML_CLUEFLOW_STYLE_NOWRAP:
 	default:
 		return "blockquote";
 	}
@@ -1249,18 +1258,17 @@ save_plain (HTMLObject *self,
 			/* FIXME we should allow wrapping on PRE sections as well */
 			len = strcspn (s, "\n");
 			
-			if (flow->style != HTML_CLUEFLOW_STYLE_PRE
-			    && flow->style != HTML_CLUEFLOW_STYLE_NOWRAP) {
+			if (flow->style != HTML_CLUEFLOW_STYLE_PRE) {
 				
-				if (unicode_strlen (s, len) > (requested_width - pad)) {
-					space = s + unicode_offset_to_index (s, requested_width - pad);
-					while (space > s 
+				if (g_utf8_strlen (s, len) > (requested_width - pad)) {
+					space = g_utf8_offset_to_pointer (s, requested_width - pad);
+					while (space 
 					       && (*space != ' '))
-						// || (IS_UTF8_NBSP ((guchar *)unicode_next_utf8 (space)))
-						// || (IS_UTF8_NBSP ((guchar *)unicode_previous_utf8 (s, space)))))
-						space = unicode_previous_utf8 (s, space);
+						/* || (IS_UTF8_NBSP ((guchar *)g_utf8_find_next_char (space, NULL)))
+						   || (IS_UTF8_NBSP ((guchar *)g_utf8_find_prev_char (s, space))))) */
+						space = g_utf8_find_prev_char (s, space);
 					
-					if (space != s)
+					if (space != NULL)
 						len = space - s;
 				}
 			}
@@ -1270,7 +1278,7 @@ save_plain (HTMLObject *self,
 			
 			/* Trim the space at the end */
 			while (*s == ' ' || IS_UTF8_NBSP (s)) 
-				s = unicode_next_utf8 (s);
+				s = g_utf8_next_char (s);
 			
 			if (*s == '\n') 
 				s++;
@@ -1298,32 +1306,37 @@ save_plain (HTMLObject *self,
 static GtkHTMLFontStyle
 get_default_font_style (const HTMLClueFlow *self)
 {
+	GtkHTMLFontStyle style = 0;
+
+	if (HTML_OBJECT (self)->parent && HTML_IS_TABLE_CELL (HTML_OBJECT (self)->parent)
+	    && HTML_TABLE_CELL (HTML_OBJECT (self)->parent)->heading)
+		style = GTK_HTML_FONT_STYLE_BOLD;
+
 	switch (self->style) {
 	case HTML_CLUEFLOW_STYLE_NORMAL:
 	case HTML_CLUEFLOW_STYLE_ITEMDOTTED:
 	case HTML_CLUEFLOW_STYLE_ITEMROMAN:
 	case HTML_CLUEFLOW_STYLE_ITEMDIGIT:
-	case HTML_CLUEFLOW_STYLE_NOWRAP:
-		return GTK_HTML_FONT_STYLE_SIZE_3;
+		return style | GTK_HTML_FONT_STYLE_SIZE_3;
 	case HTML_CLUEFLOW_STYLE_ADDRESS:
-		return GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_ITALIC;
+		return style | GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_ITALIC;
 	case HTML_CLUEFLOW_STYLE_PRE:
-		return GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED;
+		return style | GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_FIXED;
 	case HTML_CLUEFLOW_STYLE_H1:
-		return GTK_HTML_FONT_STYLE_SIZE_6 | GTK_HTML_FONT_STYLE_BOLD;
+		return style | GTK_HTML_FONT_STYLE_SIZE_6 | GTK_HTML_FONT_STYLE_BOLD;
 	case HTML_CLUEFLOW_STYLE_H2:
-		return GTK_HTML_FONT_STYLE_SIZE_5 | GTK_HTML_FONT_STYLE_BOLD;
+		return style | GTK_HTML_FONT_STYLE_SIZE_5 | GTK_HTML_FONT_STYLE_BOLD;
 	case HTML_CLUEFLOW_STYLE_H3:
-		return GTK_HTML_FONT_STYLE_SIZE_4 | GTK_HTML_FONT_STYLE_BOLD;
+		return style | GTK_HTML_FONT_STYLE_SIZE_4 | GTK_HTML_FONT_STYLE_BOLD;
 	case HTML_CLUEFLOW_STYLE_H4:
-		return GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_BOLD;
+		return style | GTK_HTML_FONT_STYLE_SIZE_3 | GTK_HTML_FONT_STYLE_BOLD;
 	case HTML_CLUEFLOW_STYLE_H5:
-		return GTK_HTML_FONT_STYLE_SIZE_2 | GTK_HTML_FONT_STYLE_BOLD;
+		return style | GTK_HTML_FONT_STYLE_SIZE_2 | GTK_HTML_FONT_STYLE_BOLD;
 	case HTML_CLUEFLOW_STYLE_H6:
-		return GTK_HTML_FONT_STYLE_SIZE_1 | GTK_HTML_FONT_STYLE_BOLD;
+		return style | GTK_HTML_FONT_STYLE_SIZE_1 | GTK_HTML_FONT_STYLE_BOLD;
 	default:
 		g_warning ("Unexpected HTMLClueFlow style %d", self->style);
-		return GTK_HTML_FONT_STYLE_DEFAULT;
+		return style | GTK_HTML_FONT_STYLE_DEFAULT;
 	}
 }
 
@@ -1636,15 +1649,14 @@ html_clueflow_init (HTMLClueFlow *clueflow,
 	object->flags &= ~HTML_OBJECT_FLAG_FIXEDWIDTH;
 
 	clue->valign = HTML_VALIGN_BOTTOM;
-	clue->halign = HTML_HALIGN_LEFT;
+	clue->halign = HTML_HALIGN_NONE;
 
 	clueflow->style = style;
 	clueflow->level = level; 
 }
 
 HTMLObject *
-html_clueflow_new (HTMLClueFlowStyle style,
-		   guint8 level)
+html_clueflow_new (HTMLClueFlowStyle style, guint8 level)
 {
 	HTMLClueFlow *clueflow;
 
@@ -1715,7 +1727,7 @@ html_clueflow_split (HTMLClueFlow *clue,
 	}
 
 	child->prev = NULL;
-	html_object_change_set (HTML_OBJECT (clue), HTML_CHANGE_ALL);
+	html_object_change_set (HTML_OBJECT (clue), HTML_CHANGE_ALL_CALC);
 
 	/* Put the children into the new clue.  */
 
@@ -1801,12 +1813,21 @@ html_clueflow_set_halignment (HTMLClueFlow *flow,
 	relayout_and_draw (HTML_OBJECT (flow), engine);
 }
 
-HTMLHAlignType
+inline HTMLHAlignType
 html_clueflow_get_halignment (HTMLClueFlow *flow)
 {
 	g_return_val_if_fail (flow != NULL, HTML_HALIGN_NONE);
 
-	return HTML_CLUE (flow)->halign;
+	if (HTML_CLUE (flow)->halign == HTML_HALIGN_NONE) {
+		if (HTML_OBJECT (flow)->parent && HTML_IS_TABLE_CELL (HTML_OBJECT (flow)->parent))
+			return HTML_CLUE (HTML_OBJECT (flow)->parent)->halign == HTML_HALIGN_NONE
+				? HTML_TABLE_CELL (HTML_OBJECT (flow)->parent)->heading ? HTML_HALIGN_CENTER : HTML_HALIGN_LEFT
+				: HTML_CLUE (HTML_OBJECT (flow)->parent)->halign;
+		else
+			return HTML_CLUE (HTML_OBJECT (flow)->parent)->halign == HTML_HALIGN_NONE
+				? HTML_HALIGN_LEFT : HTML_CLUE (HTML_OBJECT (flow)->parent)->halign;
+	} else
+		return HTML_CLUE (flow)->halign;
 }
 
 void
@@ -1986,14 +2007,13 @@ next_obj_and_clear (HTMLObject *obj, guint *off, gboolean *is_text, HTMLInterval
 static HTMLObject *
 spell_check_word_mark (HTMLObject *obj, const gchar *text, const gchar *word, guint *off, HTMLInterval *i)
 {
-	guint w_off, w_index, ioff;
-	guint len = unicode_strlen (word, -1);
+	guint w_off, ioff;
+	guint len = g_utf8_strlen (word, -1);
 	gboolean is_text;
 
 	/* printf ("[not in dictionary word off: %d off: %d]\n", word - text, *off); */
 	is_text = html_object_is_text (obj);
-	w_index = word - text;
-	w_off   = unicode_index_to_offset (text, w_index);
+	w_off   = g_utf8_pointer_to_offset (text, word);
 	while (obj && (!is_text || (is_text && *off + html_interval_get_length (i, obj) <= w_off)))
 		obj = next_obj_and_clear (obj, off, &is_text, i);
 
@@ -2008,16 +2028,16 @@ spell_check_word_mark (HTMLObject *obj, const gchar *text, const gchar *word, gu
 			ioff  = html_interval_get_start (i, obj);
 			tlen  = MIN (HTML_TEXT (obj)->text_len - toff - ioff, len);
 			t     = HTML_TEXT (obj)->text;
-			g_assert (!strncmp (text + w_index, t + unicode_offset_to_index (t, toff + ioff),
-					    unicode_offset_to_index (t, toff + ioff + tlen)
-					    - unicode_offset_to_index (t, toff + ioff)));
+			g_assert (!strncmp (word, g_utf8_offset_to_pointer (t, toff + ioff),
+					    g_utf8_offset_to_pointer (t, toff + ioff + tlen)
+					    - g_utf8_offset_to_pointer (t, toff + ioff)));
 			/* printf ("add spell error - word: %s off: %d beg: %s len: %d\n",
 			   word, *off, HTML_TEXT (obj)->text + toff, tlen); */
 			html_text_spell_errors_add (HTML_TEXT (obj),
 						    ioff + toff, tlen);
 			len     -= tlen;
 			w_off   += tlen;
-			w_index += unicode_offset_to_index (text + w_index, tlen);
+			word     = g_utf8_offset_to_pointer (word, tlen);
 			if (len)
 				do obj = next_obj_and_clear (obj, off, &is_text, i); while (obj && !is_text);
 			/* printf ("off: %d\n", *off); */
@@ -2031,11 +2051,11 @@ spell_check_word_mark (HTMLObject *obj, const gchar *text, const gchar *word, gu
 static gchar *
 begin_of_word (gchar *text, gchar *ct)
 {
-	unicode_char_t uc;
+	gunichar uc;
 
 	do
-		unicode_get_utf8 (ct, &uc);
-	while (!html_is_in_word (uc) && (ct = unicode_next_utf8 (ct)) && *ct);
+		uc = g_utf8_get_char (ct);
+	while (!html_is_in_word (uc) && (ct = g_utf8_next_char (ct)) && *ct);
 
 	return ct;
 }
@@ -2043,10 +2063,10 @@ begin_of_word (gchar *text, gchar *ct)
 static gchar *
 end_of_word (gchar *ct)
 {
-	unicode_char_t uc;
+	gunichar uc;
 	gchar *cn;
 
-	while (*ct && (cn = unicode_get_utf8 (ct, &uc)) && html_is_in_word (uc))
+	while (*ct && (cn = e_unicode_get_utf8 (ct, &uc)) && html_is_in_word (uc))
 		ct = cn;
 
 	return ct;
@@ -2068,7 +2088,8 @@ html_clueflow_spell_check (HTMLClueFlow *flow, HTMLEngine *e, HTMLInterval *inte
 	gchar *text, *ct, *word;
 	HTMLInterval *new_interval = NULL;
 
-	g_assert (HTML_OBJECT_TYPE (flow) == HTML_TYPE_CLUEFLOW);
+	g_return_if_fail (flow != NULL);
+	g_return_if_fail (HTML_OBJECT_TYPE (flow) == HTML_TYPE_CLUEFLOW);
 
 	/* if (interval)
 	   printf ("html_clueflow_spell_check %p %p %d %d\n", i->from, i->to, i->from_offset, i->to_offset); */
@@ -2110,15 +2131,15 @@ html_clueflow_spell_check (HTMLClueFlow *flow, HTMLEngine *e, HTMLInterval *inte
 				if (result == 1) {
 					gboolean is_text = (obj) ? html_object_is_text (obj) : FALSE;
 					while (obj && (!is_text
-						       || (is_text && off + html_interval_get_length (interval, obj)
-							   < unicode_index_to_offset (text, ct - text))))
+						       || (off + html_interval_get_length (interval, obj)
+							   < g_utf8_pointer_to_offset (text, ct))))
 						obj = next_obj_and_clear (obj, &off, &is_text, interval);
 				} else if (obj)
 						obj = spell_check_word_mark (obj, text, word, &off, interval);
 
 				*ct = bak;
 				if (*ct)
-					ct = unicode_next_utf8 (ct);
+					ct = g_utf8_next_char (ct);
 			}
 		}
 		g_free (text);
@@ -2145,3 +2166,50 @@ html_clueflow_is_empty (HTMLClueFlow *flow)
 		return TRUE;
 	return FALSE;
 }
+
+gint
+html_clueflow_get_line_offset (HTMLClueFlow *flow, HTMLPainter *painter, HTMLObject *child)
+{
+	HTMLObject *o, *head;
+	gint line_offset;
+
+	g_assert (HTML_IS_CLUEFLOW (flow));
+
+	if (!html_clueflow_tabs (flow, painter))
+		return -1;
+
+	line_offset = 0;
+
+	/* find head */
+	o = head = child;
+	while (o) {
+		o = head->prev;
+		if (o && o->y + o->descent - 1 < child->y - child->ascent)
+			break;
+		else
+			head = o;
+	}
+
+	if (HTML_IS_TEXT_SLAVE (head)) {
+		HTMLTextSlave *bol = HTML_TEXT_SLAVE (head);
+		line_offset = html_text_text_line_length (html_text_get_text (bol->owner, bol->posStart),
+							  0, bol->owner->text_len - bol->posStart);
+		head = html_object_next_not_slave (head);
+	}
+
+	while (head) {
+		if (head == child)
+			break;
+		line_offset += html_object_get_line_length (head, painter, line_offset);
+		head = html_object_prev_not_slave (head);
+	}
+	/* printf ("lo: %d\n", line_offset); */
+	return line_offset;
+}
+
+gboolean
+html_clueflow_tabs (HTMLClueFlow *flow, HTMLPainter *p)
+{
+	return (flow && HTML_IS_CLUEFLOW (flow) && flow->style == HTML_CLUEFLOW_STYLE_PRE) || HTML_IS_PLAIN_PAINTER (p)
+		? TRUE : FALSE;
+}
