@@ -22,6 +22,7 @@
 */
 
 #include <config.h>
+#include <stdio.h>
 #include <string.h>
 
 #include "gtkhtmldebug.h"
@@ -49,29 +50,11 @@
 
 static void        delete_object (HTMLEngine *e, HTMLObject **ret_object, guint *ret_len, HTMLUndoDirection dir);
 static void        insert_object (HTMLEngine *e, HTMLObject *obj, guint len, HTMLUndoDirection dir, gboolean check);
+static void        append_object (HTMLEngine *e, HTMLObject *o, guint len, HTMLUndoDirection dir);
+static void        insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir);
 
 /* helper functions -- need refactor */
 
-static void
-get_tree_bounds_for_merge (HTMLObject *obj, GList **first, GList **last)
-{
-	HTMLObject *cur;
-
-	* first = *last = NULL;
-
-	cur = obj;
-	while (cur) {
-		*first = g_list_append (*first, cur);
-		cur = html_object_head (cur);
-	}
-
-	cur = obj;
-	while (cur) {
-		*last = g_list_append (*last, cur);
-		cur = html_object_tail_not_slave (cur);
-	}
-}
-
 static void
 html_cursor_get_left (HTMLCursor *cursor, HTMLObject **obj, gint *off)
 {
@@ -260,8 +243,8 @@ remove_empty_and_merge (HTMLEngine *e, gboolean merge, GList *left, GList *right
 		ro  = HTML_OBJECT (right->data);
 		len = html_object_get_length (lo);
 
-		left  = left ? left->next : NULL;
-		right = right ? right->next : NULL;
+		left  = left->next;
+		right = right->next;
 
 		if (html_object_is_text (lo) && !*HTML_TEXT (lo)->text && (html_object_prev_not_slave (lo) || merge)) {
 			HTMLObject *nlo = html_object_prev_not_slave (lo);
@@ -504,19 +487,43 @@ static void
 delete_object_do (HTMLEngine *e, HTMLObject **object, guint *len)
 {
 	GList *from, *to, *left, *right;
+	gint expected_len;
 
-	if (html_engine_is_selection_active (e)) {
-		html_engine_freeze (e);
-		prepare_delete_bounds (e, &from, &to, &left, &right);
-		place_cursor_before_mark (e);
-		move_cursor_before_delete (e);
-		html_engine_disable_selection (e);
-		*len     = 0;
-		*object  = html_object_op_cut  (HTML_OBJECT (from->data), e, from->next, to->next, left, right, len);
-		remove_empty_and_merge (e, TRUE, left ? left->next : NULL, right ? right->next : NULL, NULL);
-		html_engine_spell_check_range (e, e->cursor, e->cursor);
-		html_engine_thaw (e);
-	}
+	expected_len = MAX (e->cursor->position, e->mark->position) - MIN (e->cursor->position, e->mark->position);
+	html_engine_freeze (e);
+	prepare_delete_bounds (e, &from, &to, &left, &right);
+	place_cursor_before_mark (e);
+	move_cursor_before_delete (e);
+	html_engine_disable_selection (e);
+	*len     = 0;
+	*object  = html_object_op_cut  (HTML_OBJECT (from->data), e, from->next, to->next, left, right, len);
+	remove_empty_and_merge (e, TRUE, left ? left->next : NULL, right ? right->next : NULL, NULL);
+	if (HTML_IS_TABLE (e->cursor->object) && *len && expected_len - *len)
+		html_cursor_forward_n (e->cursor, e, 1);
+	html_engine_spell_check_range (e, e->cursor, e->cursor);
+	html_engine_thaw (e);
+}
+
+static void
+check_table_0 (HTMLEngine *e)
+{
+	HTMLCursor *tail;
+
+	tail = e->mark->position < e->cursor->position ? e->cursor : e->mark;
+
+	while (tail->offset == 0 && HTML_IS_TABLE (tail->object) && e->mark->position != e->cursor->position)
+		html_cursor_backward (tail, e);
+}
+
+static void
+check_table_1 (HTMLEngine *e)
+{
+	HTMLCursor *head;
+
+	head = e->mark->position > e->cursor->position ? e->cursor : e->mark;
+
+	while (head->offset == 1 && HTML_IS_TABLE (head->object) && e->mark->position != e->cursor->position)
+		html_cursor_forward (head, e);
 }
 
 static void
@@ -527,6 +534,15 @@ delete_object (HTMLEngine *e, HTMLObject **ret_object, guint *ret_len, HTMLUndoD
 		HTMLObject *object;
 		guint len;
 
+		if (!html_clueflow_is_empty (HTML_CLUEFLOW (e->cursor->object->parent))
+		    && !html_clueflow_is_empty (HTML_CLUEFLOW (e->mark->object->parent))) {
+			check_table_0 (e);
+			check_table_1 (e);
+		}
+		if (e->cursor->position == e->mark->position) {
+			html_engine_disable_selection (e);
+			return;
+		}
 		delete_object_do (e, &object, &len);
 		if (ret_object && ret_len) {
 			*ret_object = html_object_op_copy (object, e, NULL, NULL, ret_len);
@@ -556,7 +572,20 @@ html_engine_cut (HTMLEngine *e)
  */
 
 static void
-insert_object_do (HTMLEngine *e, HTMLObject *obj, guint len, gboolean check)
+set_cursor_at_end_of_object (HTMLEngine *e, HTMLObject *o, guint len)
+{
+	guint save_position;
+
+	save_position       = e->cursor->position;
+	e->cursor->object   = html_object_get_tail_leaf (o);
+	while (html_cursor_forward (e->cursor, e))
+		;
+	e->cursor->position = save_position + len;
+	e->cursor->offset   = html_object_get_length (e->cursor->object);
+}
+
+static void
+insert_object_do (HTMLEngine *e, HTMLObject *obj, guint len, gboolean check, HTMLUndoDirection dir)
 {
 	HTMLObject *cur;
 	HTMLCursor *orig;
@@ -564,19 +593,24 @@ insert_object_do (HTMLEngine *e, HTMLObject *obj, guint len, gboolean check)
 	GList *first = NULL, *last = NULL;
 	gint level;
 
-	/* FIXME for tables */
-	if (obj->klass->type == HTML_TYPE_TABLE) {
-		html_engine_append_object (e, obj, len);
-		return;
-	}
-
 	html_engine_freeze (e);
-	if (HTML_IS_TABLE (e->cursor->object)) {
-		gint offset = e->cursor->offset;
 
-		html_engine_insert_empty_paragraph (e);
-		if (offset == 0)
-			html_cursor_backward (e->cursor, e);
+	if (HTML_IS_TABLE (e->cursor->object)) {
+		if (e->cursor->offset) {
+			HTMLObject *head = html_object_get_head_leaf (obj);
+
+			if (!head->parent || (HTML_IS_CLUEFLOW (head->parent)
+					      && !html_clueflow_is_empty (HTML_CLUEFLOW (head->parent))))
+				insert_empty_paragraph (e, dir);
+		} else {
+			HTMLObject *tail = html_object_get_tail_leaf (obj);
+
+			if (!tail->parent || (HTML_IS_CLUEFLOW (tail->parent)
+					      && !html_clueflow_is_empty (HTML_CLUEFLOW (tail->parent)))) {
+				insert_empty_paragraph (e, dir);
+				html_cursor_backward (e->cursor, e);
+			}
+		}
 	}
 
 	level = 0;
@@ -588,12 +622,10 @@ insert_object_do (HTMLEngine *e, HTMLObject *obj, guint len, gboolean check)
 	orig = html_cursor_dup (e->cursor);
 
 	html_object_change_set_down (obj, HTML_CHANGE_ALL);
-	split_and_add_empty_texts (e, level, &left, &right);
-        get_tree_bounds_for_merge (obj, &first, &last);
-
-	e->cursor->position += len;
-	e->cursor->object    = html_object_get_tail_leaf (obj);
-	e->cursor->offset    = html_object_get_length (e->cursor->object);
+	split_and_add_empty_texts (e, MIN (3, level), &left, &right);
+	first = html_object_heads_list (obj);
+	last  = html_object_tails_list (obj);
+	set_cursor_at_end_of_object (e, obj, len);
 
 	if ((left && left->data) || (right && (right->data))) {
 		HTMLObject *parent, *where;
@@ -657,8 +689,13 @@ insert_setup_undo (HTMLEngine *e, guint len, HTMLUndoDirection dir)
 static void
 insert_object (HTMLEngine *e, HTMLObject *obj, guint len, HTMLUndoDirection dir, gboolean check)
 {
-	insert_object_do (e, obj, len, check);
-	insert_setup_undo (e, len, dir);
+	/* FIXME for tables */
+	if (HTML_IS_TABLE (obj))
+		append_object (e, obj, len, dir);
+	else {
+		insert_object_do (e, obj, len, check, dir);
+		insert_setup_undo (e, len, dir);
+	}
 }
 
 void
@@ -697,8 +734,8 @@ check_magic_link (HTMLEngine *e, const gchar *text, guint len)
 		html_text_magic_link (HTML_TEXT (e->cursor->object), e, html_object_get_length (e->cursor->object));
 }
 
-void
-html_engine_insert_empty_paragraph (HTMLEngine *e)
+static void
+insert_empty_paragraph (HTMLEngine *e, HTMLUndoDirection dir)
 {
 	GList *left=NULL, *right=NULL;
 	HTMLCursor *orig;
@@ -709,7 +746,7 @@ html_engine_insert_empty_paragraph (HTMLEngine *e)
 	remove_empty_and_merge (e, FALSE, left, right, orig);
 	html_cursor_forward (e->cursor, e);
 
-	insert_setup_undo (e, 1, HTML_UNDO_UNDO);
+	insert_setup_undo (e, 1, dir);
 	g_list_free (left);
 	g_list_free (right);
 	html_engine_spell_check_range (e, orig, e->cursor);
@@ -727,6 +764,12 @@ html_engine_insert_empty_paragraph (HTMLEngine *e)
 	html_engine_insert_link (e, NULL, NULL);
 }
 
+void
+html_engine_insert_empty_paragraph (HTMLEngine *e)
+{
+	insert_empty_paragraph (e, HTML_UNDO_UNDO);
+}
+
 void
 html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
 {
@@ -734,13 +777,13 @@ html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
 	gint alen;
 
 	if (len == -1)
-		len = unicode_strlen (text, -1);
+		len = g_utf8_strlen (text, -1);
 	if (!len)
 		return;
 
 	do {
-		nl   = unicode_strchr (text, '\n');
-		alen = nl ? unicode_index_to_offset (text, nl - text) : len;
+		nl   = g_utf8_strchr (text, '\n');
+		alen = nl ? g_utf8_pointer_to_offset (text, nl) : len;
 		if (alen) {
 			HTMLObject *o;
 			gboolean check = FALSE;
@@ -763,7 +806,7 @@ html_engine_insert_text (HTMLEngine *e, const gchar *text, guint len)
 		}
 		if (nl) {
 			html_engine_insert_empty_paragraph (e);
-			len -= unicode_index_to_offset (text, nl - text) + 1;
+			len -= g_utf8_pointer_to_offset (text, nl) + 1;
 			text = nl + 1;
 		}
 	} while (nl);
@@ -783,6 +826,24 @@ html_engine_paste_text (HTMLEngine *e, const gchar *text, guint len)
 	html_undo_level_end (e->undo);
 }
 
+void
+html_engine_delete_container (HTMLEngine *e)
+{
+	g_assert (HTML_IS_ENGINE (e));
+	g_assert (e->cursor->object);
+	g_assert (html_object_is_container (e->cursor->object));
+
+	html_engine_freeze (e);
+	html_engine_set_mark (e);
+	if (e->cursor->offset)
+		html_cursor_beginning_of_line (e->cursor, e);
+	else
+		html_cursor_end_of_line (e->cursor, e);
+	html_engine_update_selection_if_necessary (e);
+	html_engine_delete (e);
+	html_engine_thaw (e);
+}
+
 void
 html_engine_delete_n (HTMLEngine *e, guint len, gboolean forward)
 {
@@ -877,8 +938,8 @@ html_engine_insert_link (HTMLEngine *e, const gchar *url, const gchar *target)
 		html_engine_set_insertion_link (e, url, target);
 }
 
-void
-html_engine_append_object (HTMLEngine *e, HTMLObject *o, guint len)
+static void
+append_object (HTMLEngine *e, HTMLObject *o, guint len, HTMLUndoDirection dir)
 {
 	GList *left = NULL, *right = NULL;
 	HTMLObject *where;
@@ -902,12 +963,20 @@ html_engine_append_object (HTMLEngine *e, HTMLObject *o, guint len)
 		html_clue_append (HTML_CLUE (flow), o);
 
 		html_object_split (e->cursor->object, e, NULL, e->cursor->offset, 2, &left, &right);
+		len += 2;
 
 		where = HTML_OBJECT (left->data);
 		html_clue_append_after (HTML_CLUE (where->parent), flow, where);
 	}
+	html_object_change_set (o, HTML_CHANGE_ALL_CALC);
+	html_engine_thaw (e);
 
 	html_cursor_forward_n (e->cursor, e, len);
-	insert_setup_undo (e, len, HTML_UNDO_UNDO);	
-	html_engine_thaw (e);
+	insert_setup_undo (e, len, dir);
+}
+
+void
+html_engine_append_object (HTMLEngine *e, HTMLObject *o, guint len)
+{
+	append_object (e, o, len, HTML_UNDO_UNDO);
 }
