@@ -74,6 +74,8 @@ struct _HTMLEngine {
            repaint everything.  */
 	guint freeze_count;
 	guint thaw_idle_id;
+	gint block_redraw;
+	gboolean need_redraw;
 
 	gboolean parsing;
 	HTMLTokenizer *ht;
@@ -115,6 +117,7 @@ struct _HTMLEngine {
 	HTMLStack *font_face_stack;
 	HTMLStack *color_stack;	/* Color stack, elements are GdkColors.  */
 	HTMLStack *clueflow_style_stack; /* Clueflow style stack, elements are HTMLClueFlowStyles.  */
+	HTMLStack *frame_stack;
 
 	gchar *url;
 	gchar *target;
@@ -127,6 +130,8 @@ struct _HTMLEngine {
 	/* timer id for parsing routine */
 	guint timerId;
 
+	guint redraw_idle_id;
+
 	/* FIXME: replace with a `gchar *'?  */
 	GString *title;
 
@@ -156,12 +161,12 @@ struct _HTMLEngine {
 	 */
 	GList *tempStrings;
 
+	HTMLMap *map;
 	HTMLForm *form;
 	HTMLSelect *formSelect;
 	HTMLTextArea *formTextArea;
 	GList *formList;
 	GString *formText;
-	gboolean noWrap;
 
 	/* This is TRUE if we cannot insert a paragraph break (which is just an
            extra empty line).  It's set to FALSE as soon as some element is
@@ -230,9 +235,16 @@ struct _HTMLEngine {
 	/* id-to-object mapping */
 	GHashTable *id_table;
 
+	/* table contains data, which are set to objects by type during parse time */
+	GHashTable *class_data;
+
 	gdouble min_split_index;
 
 	gboolean need_spell_check;
+	gint block_events;
+
+	GSList *cursor_position_stack;
+	gboolean selection_mode;
 };
 
 /* must be forward referenced *sigh* */
@@ -270,6 +282,10 @@ gboolean  html_engine_get_editable  (HTMLEngine *e);
 void  html_engine_set_focus  (HTMLEngine *engine,
 			      gboolean    have_focus);
 
+/* Tokenizer. */
+void html_engine_set_tokenizer (HTMLEngine *engine,
+				HTMLTokenizer *tok);
+
 /* Parsing control.  */
 GtkHTMLStream *html_engine_begin            (HTMLEngine  *p, 
 					     char        *content_type);
@@ -281,6 +297,7 @@ gint  html_engine_calc_min_width       (HTMLEngine *e);
 void  html_engine_calc_size            (HTMLEngine *p);
 gint  html_engine_get_doc_height       (HTMLEngine *p);
 gint  html_engine_get_doc_width        (HTMLEngine *e);
+guint html_engine_get_max_width        (HTMLEngine *e);
 void  html_engine_draw                 (HTMLEngine *e,
 					gint        x,
 					gint        y,
@@ -293,7 +310,10 @@ void  html_engine_draw_background      (HTMLEngine *e,
 					gint        height);
 
 /* Scrolling.  */
-void      html_engine_schedule_update      (HTMLEngine  *p);
+void      html_engine_schedule_update      (HTMLEngine  *e);
+void      html_engine_schedule_redraw      (HTMLEngine  *e);
+void      html_engine_block_redraw         (HTMLEngine  *e);
+void      html_engine_unblock_redraw       (HTMLEngine  *e);
 gboolean  html_engine_make_cursor_visible  (HTMLEngine  *e);
 gboolean  html_engine_goto_anchor          (HTMLEngine  *e,
 					    const gchar *anchor);
@@ -309,8 +329,7 @@ void  html_engine_queue_clear       (HTMLEngine *e,
 				     guint       height);
 
 void  html_engine_set_painter       (HTMLEngine  *e,
-				     HTMLPainter *painter,
-				     gint max_width);
+				     HTMLPainter *painter);
 
 /* Getting objects through pointer positions.  */
 HTMLObject  *html_engine_get_object_at  (HTMLEngine *e,
@@ -383,4 +402,7 @@ HTMLObject *html_engine_get_object_by_id    (HTMLEngine  *e,
 
 HTMLEngine *html_engine_get_top_html_engine (HTMLEngine *e);
 void        html_engine_thaw_idle_reset     (HTMLEngine *e);
+
+const gchar * html_engine_get_class_data (HTMLEngine *e, const gchar *class_name, const gchar *key);
+
 #endif /* _HTMLENGINE_H_ */
