@@ -18,6 +18,8 @@
     the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
     Boston, MA 02111-1307, USA.
 */
+
+#include <config.h>
 #include "htmlimageinput.h"
 #include "htmlform.h"
 #include <string.h>
@@ -186,7 +188,7 @@ html_imageinput_init (HTMLImageInput *img,
 
 	img->image = HTML_IMAGE (html_image_new (imf, 
 						 url, NULL, NULL,
-						 -1, -1, 0, 0,
+						 -1, -1, FALSE, FALSE, 0,
 						 NULL,
 						 HTML_VALIGN_BOTTOM));
 
