@@ -20,11 +20,15 @@
 */
 
 #include <config.h>
+#include "gtkhtml-compat.h"
+
 #include <gnome.h>
 #include <ctype.h>
 #include <libgnomeprint/gnome-print.h>
-#include <unicode.h>
+#include <gal/unicode/gunicode.h>
 
+#include "htmlembedded.h"
+#include "gtkhtml-embedded.h"
 #include "htmlprinter.h"
 
 /* #define PRINTER_DEBUG */
@@ -62,14 +66,14 @@ insure_paper (void)
 }
 
 static gdouble
-get_page_height (HTMLPrinter *printer)
+printer_get_page_height (HTMLPrinter *printer)
 {
 	insure_paper ();
 	return gnome_paper_psheight (paper);
 }
 
 static gdouble
-get_page_width (HTMLPrinter *printer)
+printer_get_page_width (HTMLPrinter *printer)
 {
 	insure_paper ();
 	return gnome_paper_pswidth (paper);
@@ -124,7 +128,7 @@ html_printer_coordinates_to_gnome_print (HTMLPrinter *printer,
 	print_y = SCALE_ENGINE_TO_GNOME_PRINT (engine_y);
 
 	print_x = print_x + get_lmargin (printer);
-	print_y = (get_page_height (printer) - print_y) - get_tmargin (printer);
+	print_y = (printer_get_page_height (printer) - print_y) - get_tmargin (printer);
 
 	*print_x_return = print_x;
 	*print_y_return = print_y;
@@ -349,8 +353,8 @@ draw_panel (HTMLPainter *painter,
 	gdouble height;
 
 	#define INC 0x8000
-	#define DARK(c)  dark.  ## c = MAX (((gint) bg-> ## c) - INC, 0)
-	#define LIGHT(c) light. ## c = MIN (((gint) bg-> ## c) + INC, 0xffff)
+	#define DARK(c)  dark.c = MAX (((gint) bg->c) - INC, 0)
+	#define LIGHT(c) light.c = MIN (((gint) bg->c) + INC, 0xffff)
 
 	DARK(red);
 	DARK(green);
@@ -508,8 +512,7 @@ draw_text (HTMLPainter *painter,
 	GnomeFont *font;
 	HTMLPrinter *printer;
 	gdouble print_x, print_y;
-	gchar *text_tmp;
-	gint bytes;
+	gchar *end, *text_tmp;
 
 	printer = HTML_PRINTER (painter);
 	g_return_if_fail (printer->print_context != NULL);
@@ -521,10 +524,10 @@ draw_text (HTMLPainter *painter,
 
 	/* Oh boy, this sucks so much.  The GnomePrint API could be improved to
            avoid this.  */
-	bytes = unicode_offset_to_index (text, len);
-	text_tmp = alloca (bytes + 1);
-	memcpy (text_tmp, text, bytes);
-	text_tmp [bytes] = '\0';
+	end = g_utf8_offset_to_pointer (text, len);
+	text_tmp = alloca (end - text + 1);
+	memcpy (text_tmp, text, end - text);
+	text_tmp [end - text] = '\0';
 
 	font = html_painter_get_font (painter, painter->font_face, painter->font_style);
 	gnome_print_setfont (printer->print_context, font);
@@ -569,6 +572,29 @@ draw_text (HTMLPainter *painter,
 	}
 }
 
+static void
+draw_embedded (HTMLPainter *p, HTMLEmbedded *o, gint x, gint y) 
+{
+	gdouble print_x, print_y;	
+	HTMLPrinter *printer = HTML_PRINTER(p);
+	GtkWidget *embedded_widget;
+
+	html_printer_coordinates_to_gnome_print (printer, x, y, &print_x, &print_y);
+	gnome_print_gsave(printer->print_context); 
+
+	gnome_print_translate(printer->print_context, 
+			      print_x, print_y - o->height * PIXEL_SIZE);
+ 
+	embedded_widget = html_embedded_get_widget(o);
+	if (embedded_widget && GTK_IS_HTML_EMBEDDED (embedded_widget)) {
+		gtk_signal_emit_by_name(GTK_OBJECT (embedded_widget), 
+					"draw_print", 
+					printer->print_context);
+	}
+
+	gnome_print_grestore(printer->print_context); 
+}
+
 static void
 draw_shade_line (HTMLPainter *painter,
 		 gint x, gint y,
@@ -688,6 +714,17 @@ unref_font (HTMLPainter *painter, HTMLFont *font)
 	gtk_object_unref (GTK_OBJECT (font->data));
 }
 
+static guint
+get_page_width (HTMLPainter *painter, HTMLEngine *e)
+{
+	return html_printer_get_page_width (HTML_PRINTER (painter));
+}
+
+static guint
+get_page_height (HTMLPainter *painter, HTMLEngine *e)
+{
+	return html_printer_get_page_height (HTML_PRINTER (painter));
+}
 
 static void
 init (GtkObject *object)
@@ -730,6 +767,9 @@ class_init (GtkObjectClass *object_class)
 	painter_class->draw_background = draw_background;
 	painter_class->get_pixel_size = get_pixel_size;
 	painter_class->set_clip_rectangle = set_clip_rectangle;
+	painter_class->draw_embedded = draw_embedded;
+	painter_class->get_page_width = get_page_width;
+	painter_class->get_page_height = get_page_height;
 
 	parent_class = gtk_type_class (html_painter_get_type ());
 }
@@ -781,7 +821,7 @@ html_printer_get_page_width (HTMLPrinter *printer)
 	g_return_val_if_fail (printer != NULL, 0);
 	g_return_val_if_fail (HTML_IS_PRINTER (printer), 0);
 
-	printer_width = get_page_width (printer) - get_lmargin (printer) - get_rmargin (printer);
+	printer_width = printer_get_page_width (printer) - get_lmargin (printer) - get_rmargin (printer);
 	engine_width = SCALE_GNOME_PRINT_TO_ENGINE (printer_width);
 
 	return engine_width;
@@ -796,7 +836,7 @@ html_printer_get_page_height (HTMLPrinter *printer)
 	g_return_val_if_fail (printer != NULL, 0);
 	g_return_val_if_fail (HTML_IS_PRINTER (printer), 0);
 
-	printer_height = get_page_height (printer) - get_lmargin (printer) - get_rmargin (printer);
+	printer_height = printer_get_page_height (printer) - get_lmargin (printer) - get_rmargin (printer);
 	engine_height = SCALE_GNOME_PRINT_TO_ENGINE (printer_height);
 
 	return engine_height;
